# Copyright (c) 2021 Lowatt <info@lowatt.fr>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import json
from typing import Any, Dict, Optional

import pydantic

from . import LOGGER


def grdf_json_dumps(v: Dict[str, Any], *, default: Any) -> Any:
    for key, value in v.items():
        if isinstance(value, bool):
            v[key] = "Vrai" if value else "Faux"
    return json.dumps(v, default=default)


class DeclareAccess(pydantic.BaseModel):
    pce: str
    role_tiers: str = "AUTORISE_CONTRAT_FOURNITURE"
    raison_sociale: Optional[str]
    nom_titulaire: Optional[str]
    code_postal: str
    courriel_titulaire: str
    date_consentement_declaree: str
    date_fin_autorisation_demandee: str
    perim_donnees_techniques_et_contractuelles: bool = False
    perim_historique_de_donnees: bool = False
    perim_flux_de_donnees: bool = False
    perim_donnees_informatives: bool = False
    perim_donnees_publiees: bool = False

    class Config:
        json_dumps = grdf_json_dumps
        extra = "forbid"

    @pydantic.root_validator()
    def check_oneof_nom_or_raison_sociale(
        cls,  # noqa: B902 (Invalid first argument 'cls' used for instance method.)
        values: Dict[str, str],
    ) -> Dict[str, str]:
        if not any(values.get(k) for k in ("raison_sociale", "nom_titulaire")):
            raise ValueError(
                "One of raison-sociale or nom-titulaire should be specified"
            )
        return values


class Access(pydantic.BaseModel):
    pce: str = pydantic.Field(alias="id_pce")
    etat_droit_acces: str
    perim_donnees_publiees: bool
    perim_donnees_informatives: bool
    courriel_titulaire: str
    raison_sociale_du_titulaire: Optional[str]
    nom_titulaire: Optional[str]
    statut_controle_preuve: Optional[str]

    def __str__(self) -> str:
        name = self.raison_sociale_du_titulaire or self.nom_titulaire
        return f"<PCE {self.pce} from {name} ({self.courriel_titulaire})>"

    def is_active(self) -> bool:
        """Return True if we can get data for this PCE"""
        if self.etat_droit_acces != "Active":
            LOGGER.error(
                "Could not collect data for %s: status is '%s'",
                self,
                self.etat_droit_acces,
            )
            return False
        if not any([self.perim_donnees_publiees, self.perim_donnees_informatives]):
            LOGGER.error(
                "Could not collect data for %s: both perim_donnees_publiees and "
                "perim_donnees_informatives are not set",
                self,
            )
            return False
        return True

    def check_consent(self) -> bool:
        """Return True if there's no pending/failed consent validation"""
        if self.statut_controle_preuve in ("Preuve en attente", "Preuve Vérifiée KO"):
            LOGGER.error(
                "statut_controle_preuve of %s is %r",
                self,
                self.statut_controle_preuve,
            )
            return False
        return True



# Copyright (c) 2021 Lowatt <info@lowatt.fr>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import json
import logging
import os
import sys
from typing import Any, Callable, Type

import click
import pydantic

from . import api, models

Callback = Callable[..., None]


def api_options(func: Callback) -> Callback:
    click.option(
        "--client-id",
        required="CLIENT_ID" not in os.environ,
        default=os.environ.get("CLIENT_ID"),
        help="openid client id",
    )(func)
    click.option(
        "--client-secret",
        required="CLIENT_SECRET" not in os.environ,
        default=os.environ.get("CLIENT_SECRET"),
        help="openid client id",
    )(func)
    click.option(
        "--bas",
        default=False,
        is_flag=True,
        help="Use staging (bac à sable) environment",
    )(func)
    return func


def options_from_model(
    model: Type[pydantic.BaseModel],
) -> Callable[[Callback], Callback]:
    def decorator(func: Callback) -> Callback:
        for field in reversed(list(model.__fields__.values())):
            opt = field.alias.replace("_", "-")
            opt = f"--{opt}"
            kwargs = {
                "required": field.required,
                "default": field.default,
            }
            if field.field_info.description is not None:
                kwargs["help"] = field.field_info.description
            if field.type_ == bool:
                kwargs["is_flag"] = True
            click.option(opt, **kwargs)(func)
        return func

    return decorator


@click.group()
def main() -> None:
    logging.basicConfig(level="INFO", format="%(levelname)s %(message)s")


@main.command()
@click.option(
    "--check",
    default=False,
    is_flag=True,
    help="Check droits-access and exit with error in case of consent validation error",
)
@api_options
def droits_acces(client_id: str, client_secret: str, bas: bool, check: bool) -> None:
    grdf = {True: api.StagingAPI, False: api.API}[bas](client_id, client_secret)
    if check:
        grdf.check_consent_validation()
    else:
        json.dump(grdf.droits_acces(), sys.stdout)


@main.command()
@click.argument("pce")
@click.option("--from-date", required=True)
@click.option("--to-date", required=True)
@api_options
def donnees_consos_publiees(
    client_id: str,
    client_secret: str,
    bas: bool,
    pce: str,
    from_date: str,
    to_date: str,
) -> None:
    grdf = {True: api.StagingAPI, False: api.API}[bas](client_id, client_secret)
    json.dump(grdf.donnees_consos_publiees(pce, from_date, to_date), sys.stdout)


@main.command()
@click.argument("pce")
@click.option("--from-date", required=True)
@click.option("--to-date", required=True)
@api_options
def donnees_consos_informatives(
    client_id: str,
    client_secret: str,
    bas: bool,
    pce: str,
    from_date: str,
    to_date: str,
) -> None:
    grdf = {True: api.StagingAPI, False: api.API}[bas](client_id, client_secret)
    json.dump(grdf.donnees_consos_informatives(pce, from_date, to_date), sys.stdout)


@main.command()
@click.argument("pce")
@api_options
def donnees_contractuelles(
    client_id: str,
    client_secret: str,
    bas: bool,
    pce: str,
) -> None:
    grdf = {True: api.StagingAPI, False: api.API}[bas](client_id, client_secret)
    json.dump(grdf.donnees_contractuelles(pce), sys.stdout)


@main.command()
@click.argument("pce")
@api_options
def donnees_techniques(
    client_id: str,
    client_secret: str,
    bas: bool,
    pce: str,
) -> None:
    grdf = {True: api.StagingAPI, False: api.API}[bas](client_id, client_secret)
    json.dump(grdf.donnees_techniques(pce), sys.stdout)


@main.command()
@api_options
@options_from_model(models.DeclareAccess)
def declare_acces(
    client_id: str,
    client_secret: str,
    bas: bool,
    **kwargs: Any,
) -> None:
    access = models.DeclareAccess(**kwargs)
    grdf = {True: api.StagingAPI, False: api.API}[bas](client_id, client_secret)
    grdf.declare_acces(access)



# Copyright (c) 2021 Lowatt <info@lowatt.fr>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import ndjson
import pytest
import responses

from lowatt_grdf import api, models


@pytest.fixture
def grdf():
    responses.add(
        responses.POST,
        f"{api.OPENID_ENDPOINT}/access_token",
        json={"access_token": "xxx"},
    )
    return api.API("id", "secret")


@responses.activate
def test_access_token(grdf):
    assert grdf.access_token == "xxx"


@responses.activate
def test_donnees_contractuelles(grdf):
    payload = {
        "pce": {"id_pce": "23000000000000"},
        "donnees_contractuelles": {
            "car": 9426,
            "cja": 0,
            "profil": {"profil_type_actuel": "P012"},
            "tarif_acheminement": "T2",
            "date_mes": "2020-12-07",
        },
        "statut_restitution": None,
    }
    responses.add(
        responses.GET,
        f"{grdf.api}/pce/23000000000000/donnees_contractuelles",
        json=payload,
    )
    assert grdf.donnees_contractuelles("23000000000000") == payload


@responses.activate
def test_donnees_techniques(grdf):
    payload = {
        "pce": {"id_pce": "23000000000000"},
        "donnees_techniques": {
            "situation_compteur": {
                "numero_rue": "4",
                "nom_rue": "RUE LAMBDA",
                "complement_adresse": None,
                "code_postal": "99099",
                "commune": "LAMBDAVILLE",
            },
            "caracteristiques_compteur": {
                "frequence": "1M",
                "client_sensible_mig": "Non",
            },
            "pitd": {"identifiant_pitd": "GD9999", "libelle_pitd": "DUMMYVILLE"},
        },
        "statut_restitution": None,
    }
    responses.add(
        responses.GET,
        f"{grdf.api}/pce/23000000000000/donnees_techniques",
        json=payload,
    )
    assert grdf.donnees_techniques("23000000000000") == payload


ACCESS_PAYLOAD = {
    "id_droit_acces": "2734291b-3df4-4aef-92de-ad0186e0be6b",
    "id_pce": "GI000000",
    "role_tiers": "AUTORISE_CONTRAT_FOURNITURE",
    "raison_sociale_du_tiers": "DUMMY",
    "nom_titulaire": None,
    "raison_sociale_du_titulaire": "John Doe Inc.",
    "courriel_titulaire": "jdoe@example.com",
    "code_postal": "99099",
    "perim_donnees_techniques_et_contractuelles": "Vrai",
    "perim_historique_de_donnees": "Vrai",
    "perim_flux_de_donnees": "Vrai",
    "perim_donnees_informatives": "Vrai",
    "perim_donnees_publiees": "Vrai",
    "date_creation": "2021-07-02 16:41:14",
    "etat_droit_acces": "Active",
    "date_revocation": None,
    "source_revocation": None,
    "date_passage_a_obsolete": None,
    "source_passage_a_obsolete": None,
    "date_fin_autorisation": "2023-07-02 00:00:00",
    "date_passage_a_refuse": None,
    "source_passage_a_refuse": None,
    "parcours": "CLIENT_CONNECT",
    "statut_controle_preuve": None,
    "date_limite_transmission_preuve": None,
}


@responses.activate
def test_droits_acces(grdf):
    payload = [
        ACCESS_PAYLOAD,
        {
            "code_statut_traitement": "0000000000",
            "message_retour_traitement": "L'opération s'est déroulée avec succès.",
        },
    ]
    responses.add(
        responses.GET,
        f"{grdf.api}/droits_acces",
        headers={"Content-Type": "application/nd-json"},
        body=ndjson.dumps(payload),
    )
    assert grdf.droits_acces() == payload


@responses.activate
def test_check_constent_validation_ok(grdf, caplog):
    payload = [
        ACCESS_PAYLOAD,
        {
            "code_statut_traitement": "0000000000",
            "message_retour_traitement": "L'opération s'est déroulée avec succès.",
        },
    ]
    responses.add(
        responses.GET,
        f"{grdf.api}/droits_acces",
        headers={"Content-Type": "application/nd-json"},
        body=ndjson.dumps(payload),
    )
    with caplog.at_level("INFO", logger="lowatt_grdf"):
        grdf.check_consent_validation()
    assert [r.message for r in caplog.records] == [
        "Access to <PCE GI000000 from John Doe Inc. (jdoe@example.com)> OK",
    ]


@responses.activate
def test_check_constent_validation_inactive(grdf, caplog):
    access = dict(ACCESS_PAYLOAD)
    access["etat_droit_acces"] = "À valider"
    payload = [
        access,
        {
            "code_statut_traitement": "0000000000",
            "message_retour_traitement": "L'opération s'est déroulée avec succès.",
        },
    ]
    responses.add(
        responses.GET,
        f"{grdf.api}/droits_acces",
        headers={"Content-Type": "application/nd-json"},
        body=ndjson.dumps(payload),
    )
    with caplog.at_level("INFO", logger="lowatt_grdf"):
        grdf.check_consent_validation()
    assert [r.message for r in caplog.records] == [
        "Could not collect data for <PCE GI000000 from John Doe Inc. (jdoe@example.com)>: status is 'À valider'",
    ]


@responses.activate
def test_check_constent_validation_preuve(grdf, caplog):
    access = dict(ACCESS_PAYLOAD)
    access["statut_controle_preuve"] = "Preuve en attente"
    payload = [
        access,
        {
            "code_statut_traitement": "0000000000",
            "message_retour_traitement": "L'opération s'est déroulée avec succès.",
        },
    ]
    responses.add(
        responses.GET,
        f"{grdf.api}/droits_acces",
        headers={"Content-Type": "application/nd-json"},
        body=ndjson.dumps(payload),
    )
    with caplog.at_level("INFO", logger="lowatt_grdf"), pytest.raises(
        RuntimeError, match="Theses consents have validation issues"
    ):
        grdf.check_consent_validation()
    assert [r.message for r in caplog.records] == [
        "statut_controle_preuve of <PCE GI000000 from John Doe Inc. (jdoe@example.com)> is 'Preuve en attente'",
        "Access to <PCE GI000000 from John Doe Inc. (jdoe@example.com)> OK",
    ]


@responses.activate
def test_donnees_consos_publiees(grdf):
    payload = [
        {
            "pce": {"id_pce": "23000000000000"},
            "periode": {
                "valeur": None,
                "date_debut": "2021-01-01",
                "date_fin": "2021-01-23",
            },
            "releve_debut": {
                "date_releve": None,
                "raison_releve": None,
                "libelle_raison_releve": None,
                "qualite_releve": "Mesure",
                "statut_releve": "Normal",
                "index_brut_debut": {
                    "valeur_index": 771,
                    "horodate_Index": "2021-01-01T06:00:00+01:00",
                },
                "index_converti_debut": {"valeur_index": None, "horodate_Index": None},
            },
            "releve_fin": {
                "date_releve": "2021-01-23T06:00:00+01:00",
                "raison_releve": "RNO",
                "libelle_raison_releve": "Relevé normal",
                "qualite_releve": "Mesure",
                "statut_releve": "Normal",
                "index_brut_fin": {
                    "valeur_index": 951,
                    "horodate_Index": "2021-01-23T06:00:00+01:00",
                },
                "index_converti_fin": {"valeur_index": None, "horodate_Index": None},
            },
            "consommation": {
                "date_debut_consommation": "2021-01-01T06:00:00+01:00",
                "date_fin_consommation": "2021-01-23T06:00:00+01:00",
                "flag_retour_zero": False,
                "volume_brut": 180,
                "coeff_calcul": {
                    "coeff_pta": None,
                    "valeur_pcs": None,
                    "coeff_conversion": 11.29,
                },
                "volume_converti": 0,
                "energie": 2032,
                "type_qualif_conso": "Mesuré",
                "sens_flux_gaz": "Consommation",
                "statut_conso": "Définitive",
                "journee_gaziere": None,
                "type_conso": "Publiée",
            },
            "bordereau_publication": {
                "date_debut_bordereau": None,
                "date_fin_bordereau": None,
                "nb_jour_gazier": None,
            },
            "statut_restitution": None,
        }
    ]
    responses.add(
        responses.GET,
        f"{grdf.api}/pce/23000000000000/donnees_consos_publiees",
        headers={"Content-Type": "application/nd-json"},
        body=ndjson.dumps(payload),
    )
    assert (
        grdf.donnees_consos_publiees(
            "23000000000000", from_date="2021-01-01", to_date="2021-01-23"
        )
        == payload
    )


@responses.activate
def test_donnees_consos_informatives(grdf):
    payload = [
        {
            "pce": {"id_pce": "23000000000000"},
            "periode": {
                "valeur": None,
                "date_debut": "2021-08-16",
                "date_fin": "2021-08-19",
            },
            "releve_debut": {
                "date_releve": None,
                "index_brut_debut": {"valeur_index": None, "horodate_Index": None},
                "index_converti_debut": {"valeur_index": None, "horodate_Index": None},
            },
            "releve_fin": {
                "date_releve": "2021-08-17T06:00:00+02:00",
                "index_brut_fin": {
                    "valeur_index": 1325,
                    "horodate_Index": "2021-08-17T06:00:00+02:00",
                },
                "index_converti_fin": {"valeur_index": None, "horodate_Index": None},
            },
            "consommation": {
                "date_debut_consommation": "2021-08-16T06:00:00+02:00",
                "date_fin_consommation": "2021-08-17T06:00:00+02:00",
                "flag_retour_zero": None,
                "volume_brut": 1,
                "coeff_calcul": {
                    "coeff_pta": None,
                    "valeur_pcs": None,
                    "coeff_conversion": 11.16,
                },
                "volume_converti": None,
                "energie": 8,
                "type_qualif_conso": "Mesuré",
                "sens_flux_gaz": "Consommation",
                "statut_conso": "Provisoire",
                "journee_gaziere": "2021-08-16",
                "type_conso": "Informative Journalier",
            },
            "statut_restitution": None,
        }
    ]
    responses.add(
        responses.GET,
        f"{grdf.api}/pce/23000000000000/donnees_consos_informatives",
        headers={"Content-Type": "application/nd-json"},
        body=ndjson.dumps(payload),
    )
    assert (
        grdf.donnees_consos_informatives(
            "23000000000000", from_date="2021-08-16", to_date="2021-08-17"
        )
        == payload
    )


@responses.activate
def test_declare_acces(grdf, caplog):
    access = models.DeclareAccess(
        pce="23000000000000",
        nom_titulaire="jdoe",
        code_postal="99099",
        courriel_titulaire="jdoe@example.com",
        date_consentement_declaree="2020-01-01",
        date_fin_autorisation_demandee="2023-01-01",
    )
    # XXX: use a real life response
    responses.add(responses.PUT, f"{grdf.api}/pce/23000000000000/droit_acces", json={})
    with caplog.at_level("INFO", logger="lowatt_grdf"):
        grdf.declare_acces(access)
    assert [r.message for r in caplog.records] == [
        "Successfully declared access to 23000000000000",
    ]



# The MIT License
#
# Copyright (c) 2021 Lowatt <info@lowatt.fr>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from click.testing import CliRunner

from lowatt_grdf import main


def test_cli_base() -> None:
    runner = CliRunner()
    result = runner.invoke(main.main)
    assert result.exit_code == 0
    assert (
        result.stdout
        == """Usage: main [OPTIONS] COMMAND [ARGS]...

Options:
  --help  Show this message and exit.

Commands:
  declare-acces
  donnees-consos-informatives
  donnees-consos-publiees
  donnees-contractuelles
  donnees-techniques
  droits-acces
"""
    )


def test_cli_declare_acces() -> None:
    runner = CliRunner()
    result = runner.invoke(main.main, ["declare-acces", "--help"])
    assert result.exit_code == 0
    assert (
        result.stdout
        == """Usage: main declare-acces [OPTIONS]

Options:
  --bas                           Use staging (bac à sable) environment
  --client-secret TEXT            openid client id  [required]
  --client-id TEXT                openid client id  [required]
  --pce TEXT                      [required]
  --role-tiers TEXT
  --raison-sociale TEXT
  --nom-titulaire TEXT
  --code-postal TEXT              [required]
  --courriel-titulaire TEXT       [required]
  --date-consentement-declaree TEXT
                                  [required]
  --date-fin-autorisation-demandee TEXT
                                  [required]
  --perim-donnees-techniques-et-contractuelles
  --perim-historique-de-donnees
  --perim-flux-de-donnees
  --perim-donnees-informatives
  --perim-donnees-publiees
  --help                          Show this message and exit.
"""
    )



#!/usr/bin/env python3
# Copyright (c) 2021 Lowatt <info@lowatt.fr>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from setuptools import setup


def local_scheme(version: str) -> str:
    """Generate a PEP440 compatible version if PEP440_VERSION is enabled"""
    import os
    import setuptools_scm.version  # only present during setup time

    if "PEP440_VERSION" in os.environ:
        return ""
    version = setuptools_scm.version.get_local_node_and_date(version)
    assert isinstance(version, str), version
    return version


if __name__ == "__main__":
    setup(use_scm_version={"local_scheme": local_scheme})



# Copyright (c) 2021 Lowatt <info@lowatt.fr>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import abc
import functools
import time
from typing import Any, Dict, Optional

import ndjson
import requests

from . import LOGGER, models

OPENID_ENDPOINT = "https://sofit-sso-oidc.grdf.fr/openam/oauth2/realms/externeGrdf"


def parse_grdf_bool(data: Dict[str, Any]) -> Dict[str, Any]:
    """Turn 'Vrai' and 'Faux' items from input dict into regular booleans

    >>> parse_grdf_bool({"foo": "Vrai", "bar": "Faux"})
    {'foo': True, 'bar': False}
    """
    data = dict(data)
    for key, value in data.items():
        if isinstance(value, str) and value.lower() in ("vrai", "faux"):
            data[key] = True if value.lower() == "vrai" else False
    return data


class BaseAPI(metaclass=abc.ABCMeta):
    @property
    @abc.abstractmethod
    def scope(self) -> str:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def api(self) -> str:
        raise NotImplementedError()

    def __init__(self, client_id: str, client_secret: str):
        self.client_id = client_id
        self.client_secret = client_secret
        self._access_token: Optional[str] = None
        self._last_request: Optional[float] = None

    def request(self, verb: str, *args: Any, **kwargs: Any) -> Any:
        headers = kwargs.setdefault("headers", {})
        headers["Authorization"] = f"Bearer {self.access_token}"
        if self._last_request is not None:
            time.sleep(min(max(time.time() - self._last_request, 1), 1))
        resp = requests.request(verb, *args, **kwargs)
        self._last_request = time.time()
        resp.raise_for_status()
        # XXX: duno why but without strip it doesn't work.
        # Either an implementation error of ndjson module
        # or bad response from GrDF API (at least on staging environment)
        return ndjson.loads(resp.content.strip())

    get = functools.partialmethod(request, "GET")
    post = functools.partialmethod(request, "POST")
    put = functools.partialmethod(request, "PUT")

    @property
    def access_token(self) -> str:
        if self._access_token is None:
            self._access_token = self._authenticate()
        return self._access_token

    def _authenticate(self) -> str:
        resp = requests.post(
            f"{OPENID_ENDPOINT}/access_token",
            data={
                "grant_type": "client_credentials",
                "client_id": self.client_id,
                "client_secret": self.client_secret,
                "scope": self.scope,
            },
        )
        resp.raise_for_status()
        token = resp.json()["access_token"]
        assert isinstance(token, str)
        return token

    def droits_acces(self) -> Any:
        return self.get(f"{self.api}/droits_acces")

    def check_consent_validation(self) -> None:
        resp = self.droits_acces()
        # XXX: this looks like a bug, "liste_acces" should be part of the response
        if resp and "liste_acces" in resp[0]:
            droits = resp[0]["liste_acces"]
        else:
            droits = resp
        errors = []
        for droit in droits:
            if "code_statut_traitement" in droit:
                continue
            access = models.Access(**parse_grdf_bool(droit))
            if not access.check_consent():
                errors.append(droit)
            if access.is_active():
                LOGGER.info("Access to %s OK", access)
        if errors:
            raise RuntimeError(f"Theses consents have validation issues: {errors!r}")

    def declare_acces(self, access: models.DeclareAccess) -> None:
        data = access.json(exclude={"pce"}, exclude_none=True)
        self.put(
            f"{self.api}/pce/{access.pce}/droit_acces",
            headers={"Content-Type": "application/json"},
            data=data,
        )
        LOGGER.info("Successfully declared access to %s", access.pce)

    def donnees_consos_publiees(self, pce: str, from_date: str, to_date: str) -> Any:
        return self.get(
            f"{self.api}/pce/{pce}/donnees_consos_publiees",
            params={
                "date_debut": from_date,
                "date_fin": to_date,
            },
        )

    def donnees_consos_informatives(
        self,
        pce: str,
        from_date: str,
        to_date: str,
    ) -> Any:
        return self.get(
            f"{self.api}/pce/{pce}/donnees_consos_informatives",
            params={
                "date_debut": from_date,
                "date_fin": to_date,
            },
        )

    def donnees_contractuelles(self, pce: str) -> Any:
        (payload,) = self.get(f"{self.api}/pce/{pce}/donnees_contractuelles")
        return payload

    def donnees_techniques(self, pce: str) -> Any:
        (payload,) = self.get(f"{self.api}/pce/{pce}/donnees_techniques")
        return payload


class StagingAPI(BaseAPI):
    scope = "/adict/bas/v3"
    api = "https://api.grdf.fr/adict/bas/v3"


class API(BaseAPI):
    scope = "/adict/v1"
    api = "https://api.grdf.fr/adict/v1"
